/*
 * Title:       AP_TimestampSet.h
 * Description: High-accuracy time stamp generator
 * $Copyright(c) 2008, 2009, 2011-2012 Progress Software Corporation (PSC). All rights reserved.$
 * $Copyright (c) 2013-2018 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 * @author      Henry Lockwood
 * @version     $Id: CTimeStamp.h 78362 2008-04-15 15:28:55Z hlockwoo $
 */

#ifndef AP_TIMESTAMP_SET_H
#define AP_TIMESTAMP_SET_H

#include <AP_Platform.h>
#include <IAF_Platform.h>
#include <sys/timeb.h>
#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

/** \file
 * Hires Timestamps
 */

/*
 * High-res timestamp IDs.  These are symbolic constants used to index
 * where a timestamp was created.
 */

/*
 * Downstream (incoming towards correlator) timestamp IDs
 */

/* 1000-1999 are for downstream events in the IAF */
#define AP_TIMESTAMP_DOWNSTREAM_IAF_MIN               1000

/** Optionally implemented by a transport, represents the time a message from the external system is received by the transport. */
#define AP_TIMESTAMP_DOWNSTREAM_IAF_TRANSPORT_ENTRY   1000
/** Optionally implemented by a transport, represents the time a message leaves the transport. */
#define AP_TIMESTAMP_DOWNSTREAM_IAF_TRANSPORT_EXIT    1249

/** Optionally implemented by a codec, represents the time a message from the transport enters the codec. */
#define AP_TIMESTAMP_DOWNSTREAM_IAF_DECODER_ENTRY     1250
/** Optionally implemented by a codec, represents the time a normalised event leaves the codec (for the semantic mapper). */
#define AP_TIMESTAMP_DOWNSTREAM_IAF_DECODER_EXIT      1499

/** Represents the time a normalised event enters the semantic mapper before being sent to the Correlator. */
#define AP_TIMESTAMP_DOWNSTREAM_IAF_MAPPER_ENTRY      1500
/** Represents the time a normalised event leaves the semantic mapper before being sent to the Correlator. */
#define AP_TIMESTAMP_DOWNSTREAM_IAF_MAPPER_EXIT       1749

/** Not currently implemented, do not use. */
#define AP_TIMESTAMP_DOWNSTREAM_IAF_APAMA_ENTRY       1750
/** Not currently implemented, do not use. */
#define AP_TIMESTAMP_DOWNSTREAM_IAF_APAMA_EXIT        1999
#define AP_TIMESTAMP_DOWNSTREAM_IAF_MAX               1999

/* 3000-3999 are for downstream events in the sender */
#define AP_TIMESTAMP_DOWNSTREAM_SENDER_MIN      3000
#define AP_TIMESTAMP_DOWNSTREAM_SENDER_MAX      3999

/* 5000-5999 are for downstream events in the correlator */
#define AP_TIMESTAMP_DOWNSTREAM_CORRELATOR_MIN                 5000
#define AP_TIMESTAMP_DOWNSTREAM_CORRELATOR_INPUT_MIN           5000
#define AP_TIMESTAMP_DOWNSTREAM_CORRELATOR_INPUT_MAX           5099
#define AP_TIMESTAMP_DOWNSTREAM_CORRELATOR_SERVICE_MONITOR_MIN 5100
#define AP_TIMESTAMP_DOWNSTREAM_CORRELATOR_SERVICE_MONITOR_MAX 5299
/* Reserved for internal Apama framework use */
#define AP_TIMESTAMP_DOWNSTREAM_CORRELATOR_RESERVED1_MIN       5300
#define AP_TIMESTAMP_DOWNSTREAM_CORRELATOR_RESERVED1_MAX       5499
#define AP_TIMESTAMP_DOWNSTREAM_CORRELATOR_RESERVED2_MIN       5500
#define AP_TIMESTAMP_DOWNSTREAM_CORRELATOR_RESERVED2_MAX       5699
#define AP_TIMESTAMP_DOWNSTREAM_CORRELATOR_APPLICATION_MIN     5700
#define AP_TIMESTAMP_DOWNSTREAM_CORRELATOR_APPLICATION_MAX     5899
#define AP_TIMESTAMP_DOWNSTREAM_CORRELATOR_OUTPUT_MIN          5900
#define AP_TIMESTAMP_DOWNSTREAM_CORRELATOR_OUTPUT_MAX          5999
#define AP_TIMESTAMP_DOWNSTREAM_CORRELATOR_MAX                 5999

/* 7000-7999 are for downstream events in the merger */
#define AP_TIMESTAMP_DOWNSTREAM_MERGER_MIN     7000
#define AP_TIMESTAMP_DOWNSTREAM_MERGE_MAX      7999

/* 9000-9999 are for downstream events in an external receiver client */
#define AP_TIMESTAMP_DOWNSTREAM_CLIENT_MIN      9000
#define AP_TIMESTAMP_DOWNSTREAM_CLIENT_MAX      9999

/*
 * Upstream (outgoing from correlator) timestamp IDs
 */

/* 11000-11999 are for upstream events in an external sender client */
#define AP_TIMESTAMP_UPSTREAM_CLIENT_MIN      11000
#define AP_TIMESTAMP_UPSTREAM_CLIENT_MAX      11999

/* 13000-13999 are for upstream events in the sender */
#define AP_TIMESTAMP_UPSTREAM_SENDER_MIN      13000
#define AP_TIMESTAMP_UPSTREAM_SENDER_MAX      13999

/* 15000-15999 are for upstream events in the correlator */
#define AP_TIMESTAMP_UPSTREAM_CORRELATOR_MIN                   15000
#define AP_TIMESTAMP_UPSTREAM_CORRELATOR_INPUT_MIN             15000
#define AP_TIMESTAMP_UPSTREAM_CORRELATOR_INPUT_MAX             15099
#define AP_TIMESTAMP_UPSTREAM_CORRELATOR_APPLICATION_MIN       15100
#define AP_TIMESTAMP_UPSTREAM_CORRELATOR_APPLICATION_MAX       15299
/* Reserved for internal Apama framework use */
#define AP_TIMESTAMP_UPSTREAM_CORRELATOR_RESERVED2_MIN         15300
#define AP_TIMESTAMP_UPSTREAM_CORRELATOR_RESERVED2_MAX         15499
#define AP_TIMESTAMP_UPSTREAM_CORRELATOR_RESERVED1_MIN         15500
#define AP_TIMESTAMP_UPSTREAM_CORRELATOR_RESERVED1_MAX         15699
#define AP_TIMESTAMP_UPSTREAM_CORRELATOR_SERVICE_MONITOR_MIN   15700
#define AP_TIMESTAMP_UPSTREAM_CORRELATOR_SERVICE_MONITOR_MAX   15899
#define AP_TIMESTAMP_UPSTREAM_CORRELATOR_OUTPUT_MIN            15900
#define AP_TIMESTAMP_UPSTREAM_CORRELATOR_OUTPUT_MAX            15999
#define AP_TIMESTAMP_UPSTREAM_CORRELATOR_MAX                   15999

/* 17000-17999 are for upstream events in the merger */
#define AP_TIMESTAMP_UPSTREAM_MERGER_MIN      17000
#define AP_TIMESTAMP_UPSTREAM_MERGER_MAX      17999

/* 19000-19999 are for upstream events in the IAF */
#define AP_TIMESTAMP_UPSTREAM_IAF_MIN               19000
/** Not currently implemented, do not use. */
#define AP_TIMESTAMP_UPSTREAM_IAF_APAMA_ENTRY       19000
/** Not currently implemented, do not use. */
#define AP_TIMESTAMP_UPSTREAM_IAF_APAMA_EXIT        19249

/** Represents the time an Apama event from the Correlator enters the semantic mapper. */
#define AP_TIMESTAMP_UPSTREAM_IAF_MAPPER_ENTRY      19250
/** Represents the time an Apama event from the Correlator leaves the semantic mapper. */
#define AP_TIMESTAMP_UPSTREAM_IAF_MAPPER_EXIT       19499

/** Optionally implemented by a codec, represents the time a normalised event from the semantic mapper enters the codec. */
#define AP_TIMESTAMP_UPSTREAM_IAF_ENCODER_ENTRY     19500
/** Optionally implemented by a codec, represents the time a normalised event leaves the codec. */
#define AP_TIMESTAMP_UPSTREAM_IAF_ENCODER_EXIT      19749

/** Optionally implemented by a transport, represents the time a message from the codec enters the transport. */
#define AP_TIMESTAMP_UPSTREAM_IAF_TRANSPORT_ENTRY   19750
/** Optionally implemented by a transport, represents the time a message leaves the transport and is to sent to the external system. */
#define AP_TIMESTAMP_UPSTREAM_IAF_TRANSPORT_EXIT    19999

#define AP_TIMESTAMP_UPSTREAM_IAF_MAX               19999


/**
 * Implementation of the timestamp itself.  This is a collection of index-time
 * pairs; indices must be unique within each timestamp.
 */

/**
 * Forward declare some classes
 */
struct AP_TimestampSet;

/**
 * Field keys are integer indices.  These are referred to by symbolic constants, 
 * defined above.
 */
typedef AP_uint64 AP_TimestampSetIndex;


/**
 * Field values are double-precision floating point numbers.  These measure time in
 * seconds.
 */
typedef double AP_TimestampSetTime;


/**
 * AP_TimestampSet_Functions
 *
 * Table of client visible functions exported by a timestamp instance.
 * These are the only operations that may be performed by users of the timestamp
 * object.
 *
 * Note that all of these functions take an initial AP_TimestampSet*
 * argument; this is analogous to the (hidden) 'this' pointer passed to a C++
 * object when a member function is invoked on it.
 */

struct AP_TimestampSet_Functions 
{
	/**
	 * addNow
	 *
	 * Add a new index-time pair to the timestamp.  The given index and current
	 * time will be copied into the timestamp.  If an element with the same index
	 * already exists, it will NOT be overwritten.
	 *
	 * @param timestamp The AP_TimestampSet instance
	 * @param index The index of the new element
	 */
	void (*addNow)(struct AP_TimestampSet* timestamp, AP_TimestampSetIndex index);

	/**
	 * addTime
	 *
	 * Add a new index-time pair to the timestamp.  The given index and time will
	 * be copied into the timestamp.  If an element with the same index already
	 * exists, it will NOT be overwritten.
	 *
	 * @param timestamp The AP_TimestampSet instance
	 * @param index The index of the new element
	 * @param time The time of the new element
	 */
	void (*addTime)(struct AP_TimestampSet* timestamp, AP_TimestampSetIndex index, AP_TimestampSetTime time);

	/**
	 * replace
	 *
	 * Change the time of an indexed element in the timestamp.  If an
	 * element with the given index already exists, its time will be replaced
	 * with a copy of the given time.  Otherwise, a new element is created
	 * just as though addTime() had been called.
	 *
	 * @param timestamp The AP_TimestampSet instance
	 * @param index The index of the element to be updated
	 * @param time The new time for the indexed element
	 */
	void (*replace)(struct AP_TimestampSet* timestamp, AP_TimestampSetIndex index, AP_TimestampSetTime newTime);

	/**
	 * replaceWithNow
	 *
	 * Change the time of an indexed element in the timestamp.  If an
	 * element with the given index already exists, its time will be replaced
	 * with a copy of the given time.  Otherwise, a new element is created
	 * just as though addNow() had been called.
	 *
	 * @param timestamp The AP_TimestampSet instance
	 * @param index The index of the element to be updated
	 */
	void (*replaceWithNow)(struct AP_TimestampSet* timestamp, AP_TimestampSetIndex index);

	/**
	 * findTime
	 *
	 * Search for an indexed element in the timestamp and return its
	 * time.
	 *
	 * @param timestamp The AP_TimestampSet instance
	 * @param index The element index to search for
	 * @return The time associated with the given index, or 0.0 if no
	 * matching element could be found.
	 */
	AP_TimestampSetTime (*findTime)(struct AP_TimestampSet* timestamp, AP_TimestampSetIndex index);

	/**
	 * findTimeAndRemove
	 *
	 * Search for an indexed element in the timestamp and return its
	 * time.  If found, the element will also be removed from the timestamp.
	 *
	 * @param timestamp The AP_TimestampSet instance
	 * @param index The element index to search for
	 * @return The time associated with the given index, or NULL if no
	 * matching element could be found.  If the element was found, it will be
	 * deleted from the timestamp.  Note that this function cannot distinguish
	 * between an element with a missing time and an element that does not
	 * exist - NULL will be returned in either case. The returned object 
	 * (if any) must be explicitly deleted by the caller after use. 
	 */
	AP_TimestampSetTime (*findTimeAndRemove)(struct AP_TimestampSet* timestamp, AP_TimestampSetIndex index);

	/**
	 * getNextStampAndRemove
	 */
	AP_bool (*getNextStampAndRemove)(struct AP_TimestampSet* timestamp, AP_TimestampSetIndex* retIndex, AP_TimestampSetTime* retTime);

	/**
	 * getSize
	 *
	 * Return the number of elements in the AP_TimestampSet
	 *
	 * @param timestamp The AP_TimestampSet instance
	 * @return The number of elements in the timestamp set
	 */
	AP_uint32 (*getSize)(struct AP_TimestampSet* timestamp);

	/**
	 * toString
	 *
	 * Return a printable string representation of the timestamp.  The
	 * returned string is owned by the caller and should be freed when it is
	 * no longer required using the char8free function.
	 *
	 * @param timestamp The AP_TimestampSet instance
	 * @return Printable string representation of the timestamp
	 */
	AP_char8* (*toString)(struct AP_TimestampSet* timestamp);
};


/**
 * AP_TimestampSet
 *
 * External (client-visible) interface to a timestamp object.  The
 * AP_TimestampSet struct contains a table of function pointers, declared
 * in the AP_TimestampSet_Functions struct above.  The implementation of
 * these functions is private.  Users of the timestamp should invoke functions on
 * it as in the following example (timestamp is of type AP_TimestampSet*):
 *
 * it = timestamp->functions->first(timestamp);
 */
struct AP_TimestampSet {

	/**
	 * Pointer to private internal data.
	 */
	void* reserved;

	/**
	 * Function table of internal operations on the timestamp object.  See
	 * documentation of AP_TimestampSet_Functions for details.
	 */
	struct AP_TimestampSet_Functions* functions;

};
typedef struct AP_TimestampSet AP_TimestampSet;


/**
 * AP_TimestampSet_ctor
 *
 * Constructs a new timestamp instance.
 *
 * @return Pointer to a new timestamp instance, or NULL if the new
 * instance could not be created for some reason.
 */
AP_COMMON_API AP_TimestampSet* AP_TimestampSet_ctor();


/**
 * AP_TimestampSet_copy
 *
 * Creates a deep copy of an existing timestamp instance.
 *
 * @param old The existing timestamp to be copied.
 * @return Pointer to a new timestamp instance whose contents are an
 * exact copy of old, or NULL if the new instance could not be created for
 * some reason.
 */
AP_COMMON_API AP_TimestampSet* AP_TimestampSet_copy(AP_TimestampSet* old);


/**
 * AP_TimestampSet_dtor
 *
 * Destroys a timestamp object.  This function should be called as soon
 * as the timestamp is no longer required, to free any resources it occupies.
 *
 * @param timestamp The timestamp to be destroyed.
 */
AP_COMMON_API void AP_TimestampSet_dtor(AP_TimestampSet* timestamp);

/**
 * AP_TimestampSet_getRelativeTime
 *
 * Get a current high-resolution timestamp in a format suitable for passing
 * to other AP_TimestampSet functions.  The timestamp value represents the
 * number of seconds that have passed since some (unspecified) epoch time.
 * Timestamps return by this function have no necessary relationship to wall
 * clock time and should only be used to compare against other "relative"
 * timestamps from the same machine.
 *
 * @return Current high-resolution relative timestamp
 */
AP_COMMON_API AP_TimestampSetTime AP_TimestampSet_getRelativeTime();

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* __CTIMESTAMP_HA_INCLUDED__ */
