/*
 * EventCodec.h
 *
 * External (client-visible) interface to event encoding/decoding (codec)
 * plugin libraries within the Apama Integration Adapter Framework (IAF).
 * This is the interface that must be implemented by all codec libraries in
 * order to be successfully loaded and integrated by the IAF.
 *
 * $Copyright(c) 2002-2007, 2008 Progress Software Corporation (PSC). All rights reserved.$
 * $Copyright (c) 2013, 2016-2018 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 *
 * $Revision: 340684 $
 */

#ifndef AP_EVENT_CODEC_H
#define AP_EVENT_CODEC_H


#include <AP_Types.h>
#include <IAF_Platform.h>
#include <NormalisedEvent.h>
#include <EventTransport.h>
#include <SemanticMapper.h>
#include <AP_TimestampSet.h>
#include <IAF_TimestampConfig.h>

/** \file 
 * EventCodec.
 * External (client-visible) interface to event encoding/decoding (codec)
 * plugin libraries within the Apama Integration Adapter Framework (IAF).
 * This is the interface that must be implemented by all codec libraries in
 * order to be successfully loaded and integrated by the IAF.
 */

/**
 * Codec API version.  The upper 16 bits identify the major release of the
 * API; the lower 16 bits indicate minor versions within that release.  No
 * changes that break backward binary compatibility will be made within a
 * given major release of the API.  That is, minor releases may add
 * functionality and fix bugs, but will not change or remove any documented
 * behaviour.
 *
 * Version history:
 * 1.0	  First public version, released with Apama 2.2.0
 * 1.1	  Add versioning and multiple transport support for Apama 2.2.1
 * 2.0	  Add char8free method to the NormalisedEvent functions
 *		  Remove deprecated setEventTransport() function
 * 4.0	  Addition of high-accuracy timestamps for latency measurement.
 */
#define AP_EVENTCODEC_VERSION	  0x00040000


/**
 * Codec capabilities bits.  Used in the 'info' function exported from the
 * codec library.
 */
#define AP_EVENTCODEC_CAP_ENCODER  0x0001
#define AP_EVENTCODEC_CAP_DECODER  0x0002


/**
 * Names of functions exported from shared library.  The IAF will search for
 * these functions by name when the library is loaded, so you must use these
 * *exact* names when implementing a codec library.
 */
#define AP_EVENTCODEC_INFO_FUNCTION_NAME AP_EventCodec_info
#define AP_EVENTCODEC_CTOR_FUNCTION_NAME AP_EventCodec_ctor
#define AP_EVENTCODEC_DTOR_FUNCTION_NAME AP_EventCodec_dtor


#ifdef __cplusplus
extern "C" {
#endif  /* __cplusplus */


/*
 * Forward-declare some structures
 */
struct AP_EventCodec;
struct AP_EventEncoder;
struct AP_EventDecoder;
struct AP_EventTransport;


/**
 * AP_EventCodecError
 *
 * Error codes that can be returned by codec library functions.  The
 * enumeration values follow the normal UNIX convention of zero == OK and
 * non-zero == error.
 */
typedef enum {
	/// Everything is fine 
	AP_EventCodec_OK = 0,		   	
	/// Some unspecified internal error occurred 
	AP_EventCodec_InternalError,	
	/// Couldn't encode an incoming normalised event 
	AP_EventCodec_EncodingFailure,  
	/// Couldn't decode an incoming customer event 
	AP_EventCodec_DecodingFailure,  
	/// Trouble sending encoded event to transport 
	AP_EventCodec_TransportFailure, 
	/// Trouble sending decoded event to Semantic Mapper 
	AP_EventCodec_MappingFailure,   
	/// Codec was passed an invalid property set 
	AP_EventCodec_BadProperties	 	
} AP_EventCodecError;


/**
 * AP_EventCodecProperty
 *
 * A single codec property, corresponding to a &lt;property&gt; element in the
 * adapter configuration file.
 */
typedef struct {
	/// property name 
	AP_char8* name;
	/// property value
	AP_char8* value;
} AP_EventCodecProperty;



/**
 * AP_EventCodecProperties
 *
 * Properties for the codec, extracted from the &lt;codec&gt; element in the adapter
 * configuration file.
 */
typedef struct {
	/// properties name 
	AP_char8* name;
	/// array of codec property values 
	AP_EventCodecProperty** properties;
} AP_EventCodecProperties;


/**
 * AP_EventCodecStatus
 *
 * Codec status information structure, filled in by the getStatus call.
 *
 * The first field (status) is a free-form text string that the codec 
 * can use to report any custom status information. Note that the length
 * of status is limited to 1024 characters. Longer strings will be 
 * silently truncated. The next two fields (totalEncoded and totalDecoded)
 * report the total number of events processed. The final field can contain 
 * custom information such as the state of the codec. 
 */
typedef struct {
	/**
	 * Free form text string containing a codec specific status message. 
	 * Strings longer than 1024 will be truncated
	 */
	AP_char8* status;
	/**
	 * Total number of downstream events decoded since IAF was run.
	 */
	AP_uint64 totalDecoded;
	/**
	 * Total number of encoded events since the IAF was run
	 */
	AP_uint64 totalEncoded;
	/**
	 * Additional status information about this codec.
	 * 
	 * Standard  statusDictionary keys are 
 	 * <pre><code>
	 * VERSION=codec_version_string
	 * CONFIG_VERSION=config_version_string
	 * CONNECTION=connectionGeneration (If the adapter manages only one connection)
	 * or
	 * CONNECTION_connectionId=connectionGeneration (If the adpater manages multiple connections)
 	 * </code></pre>
	 */
	AP_NormalisedEvent* statusDictionary;
} AP_EventCodecStatus;


/**
 * AP_EventCodec_Functions
 *
 * Table of client visible functions exported by a codec library instance.
 * These functions declare the only operations that may be performed by users
 * of a codec (but note that separate function tables exist for encoding- and
 * decoding-specific functions).
 *
 * Note that all of these functions take an initial AP_EventCodec* argument;
 * this is analogous to the (hidden) 'this' pointer passed to a C++ object
 * when a member function is invoked on it.
 */
struct AP_EventCodec_Functions {

	/**
	 * updateProperties
	 *
	 * Update the configuration of the codec.  The codec may assume that
	 * flushUpstream() and flushDownstream() have been called before this
	 * function is invoked.  The recommended procedure for updating properties
	 * is to first compare the new property set with the existing stored
	 * properties -- if there are no changes, no action should be taken.  Any
	 * pointer to the old property set becomes invalid as soon as this
	 * function returns; any such pointers should therefore be discarded in
	 * favour of the supplied new properties.
	 *
	 * @param codec The event codec instance
	 * @param properties The new codec property set derived from the IAF
	 * configuration file
	 * @param timestampConfig Timestamp recording/logging settings
	 * @return Event codec error code.  If this is not AP_EventCodec_OK, the
	 * getLastError() function of the codec should be called to get a more
	 * detailed description of what went wrong.
	 */
	AP_EventCodecError (*updateProperties)(struct AP_EventCodec* codec, AP_EventCodecProperties* properties, IAF_TimestampConfig* timestampConfig);

	/**
	 * getLastError
	 *
	 * Return the codec's stored error message, if any.  The message string is
	 * owned by the codec so should not be modified or freed by the caller.
	 *
	 * @param codec The event codec instance
	 * @return The last error message generated by the codec
	 */
	const AP_char8* (*getLastError)(struct AP_EventCodec* codec);

	/**
	 * getStatus
	 *
	 * Fill in the supplied AP_EventCodecStatus structure with up-to-date
	 * status information for the codec.  Note that any data pointed to by the
	 * returned structure (such as strings) remains owned by the codec.  The
	 * caller must copy this data if it wishes to modify it.
	 *
	 * @param codec The event codec instance
	 * @param status The status structure to be filled in
	 */
	void (*getStatus)(struct AP_EventCodec* codec, AP_EventCodecStatus* status);
};


/**
 * AP_EventEncoder_Functions
 *
 * Table of client visible functions exported by the encoder part of a codec
 * library instance.  These functions declare the only operations that may be
 * performed by users of an encoder.
 *
 * Note that all of these functions take an initial AP_EventEncoder* argument;
 * this is analogous to the (hidden) 'this' pointer passed to a C++ object
 * when a member function is invoked on it.
 */
struct AP_EventEncoder_Functions {

	/**
	 * sendNormalisedEvent
	 *
	 * Called by the Semantic Mapper to encode an event and send it on to the
	 * event transport.
	 *
	 * @param encoder The event encoder instance
	 * @param event The event to be encoded. Ownership is transferred to the
	 * callee.
	 * @param timeStamp Timestamps associated with this event. Ownership is
	 * transferred to the callee.
	 * @return Event codec error code.  If this is not AP_EventCodec_OK, the
	 * getLastError() function of the encoder should be called to get a more
	 * detailed description of what went wrong.
	 */
	AP_EventCodecError (*sendNormalisedEvent)(struct AP_EventEncoder* encoder, AP_NormalisedEvent* event, AP_TimestampSet* timeStamp);

	/**
	 * flushUpstream
	 *
	 * Flush any pending normalised events into the attached event transport.
	 * If event processing in the encoder is synchronous (as it usually will
	 * be) this function need not do anything except return AP_EventCodec_OK.
	 *
	 * @param encoder The event encoder instance
	 * @return Event codec error code.  If this is not AP_EventCodec_OK, the
	 * getLastError() function of the encoder should be called to get a more
	 * detailed description of what went wrong.
	 */
	AP_EventCodecError (*flushUpstream)(struct AP_EventEncoder* encoder);

	/**
	 * getLastError
	 *
	 * Return the encoder's stored error message, if any.  The message string
	 * is owned by the encoder so should not be modified or freed by the
	 * caller.
	 *
	 * @param encoder The event encoder instance
	 * @return The last error message generated by the encoder
	 */
	const AP_char8* (*getLastError)(struct AP_EventEncoder* encoder);

	/**
	 * addEventTransport
	 *
	 * Add a named event transport to the set of transports known to the
	 * encoder.
	 *
	 * If the named transport already exists, it should be replaced.
	 *
	 * @param encoder The event encoder instance
	 * @param name The name of the transport to be added.  
	 * @param transport The transport object itself
	 */
	void (*addEventTransport)(struct AP_EventEncoder* encoder, const AP_char8* name, struct AP_EventTransport* transport);

	/**
	 * removeEventTransport
	 *
	 * Remove a named event transport from the set of transports known to the
	 * encoder.
	 *
	 * If the named transport does not exist, the function should do nothing.
	 *
	 * @param encoder The event encoder instance
	 * @param name The name of the transport to be removed
	 */
	void (*removeEventTransport)(struct AP_EventEncoder* encoder, const AP_char8* name);
};


/**
 * AP_EventDecoder_Functions
 *
 * Table of client visible functions exported by the decoder part of a codec
 * library instance. These functions declare the only operations that may be
 * performed by users of a decoder.
 *
 * Note that all of these functions take an initial AP_EventDecoder* argument;
 * this is analogous to the (hidden) 'this' pointer passed to a C++ object
 * when a member function is invoked on it.
 */
struct AP_EventDecoder_Functions {

	/**
	 * sendTransportEvent
	 *
	 * Called by the event transport to decode an event and send it on to
	 * the Semantic Mapper. It is assumed that the encoder and transport
	 * share the same definition of the content of the event, so that the
	 * transport can effectively interpret the event and free any
	 * dynamically-allocated memory.
	 *
	 * @param decoder The event decoder instance
	 * @param event The event to be decoded. Ownership is transferred to the
	 * callee.
	 * @param timeStamp Timestamps associated with this event. Ownership is
	 * transferred to the callee.
	 * @return Event codec error code.  If this is not AP_EventCodec_OK, the
	 * getLastError() function of the decoder should be called to get a more
	 * detailed description of what went wrong.
	 */
	AP_EventCodecError (*sendTransportEvent)(struct AP_EventDecoder* decoder, AP_TransportEvent event, AP_TimestampSet* timeStamp);

	/**
	 * setSemanticMapper
	 *
	 * Set the Semantic Mapper object to be used by the decoder.  Currently
	 * only a single Semantic Mapper is supported in each adapter instance.
	 *
	 * @param decoder The event decoder instance
	 * @param mapper The Semantic Mapper to be used
	 */
	void (*setSemanticMapper)(struct AP_EventDecoder* decoder, AP_SemanticMapper* mapper);

	/**
	 * flushDownstream
	 *
	 * Flush any pending transport events into the attached Semantic Mapper.
	 * If event processing in the decoder is synchronous (as it usually will
	 * be) this function need not do anything except return AP_EventCodec_OK.
	 *
	 * @param decoder The event decoder instance
	 * @return Event codec error code.  If this is not AP_EventCodec_OK, the
	 * getLastError() function of the decoder should be called to get a more
	 * detailed description of what went wrong.
	 */
	AP_EventCodecError (*flushDownstream)(struct AP_EventDecoder* decoder);

	/**
	 * getLastError
	 *
	 * Return the decoder's stored error message, if any.  The message string
	 * is owned by the decoder so should not be modified or freed by the
	 * caller.
	 *
	 * @param decoder The event decoder instance
	 * @return The last error message generated by the decoder
	 */
	const AP_char8* (*getLastError)(struct AP_EventDecoder* decoder);
};


/**
 * AP_EventEncoder
 *
 * External (client-visible) interface to the encoder part of a codec plugin
 * library.  The AP_EventEncoder struct contains a table of function pointers,
 * declared in the AP_EventEncoder_Functions struct above.  The implementation
 * of these functions is private.  Users of the encoder should invoke
 * functions on it as in the following example (encoder is of type
 * AP_EventEncoder*):
 *
 * i = encoder->functions->sendNormalisedEvent(encoder, event, timestamps);
 */
struct AP_EventEncoder {

	/**
	 * Pointer to private internal data.
	 */
	void* reserved;

	/**
	 * Function table of encoder operations.  See documentation for
	 * AP_EventEncoder_Functions for details.
	 */
	struct AP_EventEncoder_Functions* functions;
};
typedef struct AP_EventEncoder AP_EventEncoder;


/**
 * AP_EventDecoder
 *
 * External (client-visible) interface to the decoder part of a codec plugin
 * library.  The AP_EventDecoder struct contains a table of function pointers,
 * declared in the AP_EventEDecoder_Functions struct above.  The implementation
 * of these functions is private.  Users of the decoder should invoke
 * functions on it as in the following example (decoder is of type
 * AP_EventDecoder*):
 *
 * i = decoder->functions->sendTransportEvent(decoder, event, timestamps);
 */
struct AP_EventDecoder {

	/**
	 * Pointer to private internal data.
	 */
	void* reserved;

	/**
	 * Function table of decoder operations.  See documentation for
	 * AP_EventDecoder_Functions for details.
	 */
	struct AP_EventDecoder_Functions* functions;
};
typedef struct AP_EventDecoder AP_EventDecoder;


/**
 * AP_EventCodec
 *
 * External (client-visible) interface to an IAF codec plugin library.  The
 * AP_EventCodec struct contains pointers to the encoder and decoder parts of
 * the codec (either of these may be NULL if the codec operates in one
 * direction only) and a table of function pointers, declared in the
 * AP_EventCodec_Functions struct above.  The implementation of these
 * functions is private.  Users of the codec should invoke functions on it as
 * in the following example (codec is of type AP_EventCodec*):
 *
 * i = codec->functions->updateProperties(codec, properties, timestampConfig);
 */
struct AP_EventCodec {

	/**
	 * Pointer to private internal data.
	 */
	void* reserved;

	/**
	 * Function table of codec operations.  See documentation for
	 * AP_EventCodec_Functions for details.
	 */
	struct AP_EventCodec_Functions* functions;

	/**
	 * Pointer to embedded encoder.
	 */
	AP_EventEncoder* encoder;

	/**
	 * Pointer to embedded decoder.
	 */
	AP_EventDecoder* decoder;
};
typedef struct AP_EventCodec AP_EventCodec;


/**
 * AP_EventCodecInfoPtr
 *
 * Pointer to the info function for the codec library.  Each codec plugin
 * library must export a function with this signature, named using the
 * AP_EVENTCODEC_INFO_FUNCTION_NAME macro.
 *
 * Returns basic information about the codecs implemented by this shared
 * library, specifically whether they are encoders, decoders or bidirectional
 * codecs.
 *
 * @param version 'Out' parameter for version number of codec library.  This
 * should be the value of AP_EVENTCODEC_VERSION that was defined in
 * EventCodec.h when the codec was built.  The IAF will check the version
 * number to ensure that it can support all the capabilities offered by the
 * codec.
 * @param capabilities 'Out' parameter for the capabilities of the codecs
 * provided by this library.  This should be the sum of the appropriate
 * AP_EVENTCODEC_CAP_* constants found in EventCodec.h.
 */
typedef AP_EVENTCODEC_API void (AP_EVENTCODEC_CALL* AP_EventCodecInfoPtr)(AP_uint32* version, AP_uint32* capabilities);


/**
 * AP_EventCodecCtorPtr
 *
 * Pointer to the constructor function for the codec library.  Each codec
 * plugin library must export a function with this signature, named using the
 * AP_EVENTCODEC_CTOR_FUNCTION_NAME macro.
 *
 * Constructs a new instance of the event codec.  This function will be called
 * by the adapter main program when the adapter starts up.
 *
 * Note that the input parameters (name, properties, timestampConfig) are owned
 * by the caller.  The IAF framework guarantees that properties and
 * timestampConfig will remain valid until after a subsequent call to
 * updateProperties(), so it is safe to hold a pointer to this structure.
 * You should, however, copy the name string if you wish to keep it.  The
 * contents of the output parameter errMsg belongs to the codec.
 *
 * @param name The name of this codec instance, also found in properties
 * @param properties Codec property set derived from the IAF config file
 * @param err 'Out' parameter for error code if constructor fails
 * @param errMsg 'Out' parameter for error message if constructor fails
 * @param timestampConfig Timestamp recording/logging settings
 * @return Pointer to a new codec instance or NULL if the constructor fails
 * for some reason.  In the case of failure, err and errMsg should be filled
 * in with an appropriate error code and message describing the failure.
 */
typedef AP_EVENTCODEC_API AP_EventCodec* (AP_EVENTCODEC_CALL* AP_EventCodecCtorPtr)(AP_char8* name, AP_EventCodecProperties* properties, AP_EventCodecError* err, AP_char8** errMsg, IAF_TimestampConfig* timestampConfig);


/**
 * AP_EventCodecDtorPtr
 *
 * Pointer to the destructor function for the codec library.  Each codec
 * plugin library must export a function with this signature, named using the
 * AP_EVENTCODEC_DTOR_FUNCTION_NAME macro.
 *
 * Destroys an instance of the event codec that was previously created by
 * AP_EventCodec_ctor.  The codec may assume that it has been flushed before
 * the destructor is called.
 *
 * @param codec The event codec object to be destroyed
 */
typedef AP_EVENTCODEC_API void (AP_EVENTCODEC_CALL* AP_EventCodecDtorPtr)(AP_EventCodec* codec);


#ifdef __cplusplus
}  /* extern "C" */
#endif  /* __cplusplus */


#endif  /* AP_EVENT_CODEC_H */
