/*
 * SemanticMapper.h
 *
 * External (client-visible) interface to the Semantic Mapper component of the
 * Apama Integration Adapter Framework (IAF).
 *
 * $Copyright(c) 2002, 2004-2006, 2008-2009 Progress Software Corporation (PSC). All rights reserved.$
 * $Copyright (c) 2013, 2016-2018 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 *
 * $Id: SemanticMapper.h 339852 2018-10-24 14:12:31Z jhea $
 */

#ifndef AP_SEMANTIC_MAPPER_H
#define AP_SEMANTIC_MAPPER_H

#include <AP_Types.h>
#include <NormalisedEvent.h>
#include <AP_TimestampSet.h>

/** \file 
 * Semantic Mapper
 */

#ifdef __cplusplus
extern "C" {
#endif  /* __cplusplus */


/*
 * Forward declare some structures
 */
struct AP_SemanticMapper;


/**
 * AP_SemanticMapperError
 *
 * Error codes that can be returned by operations on the Semantic Mapper.  The
 * enumeration values follow the normal UNIX convention of zero == OK and
 * non-zero == error.
 */
typedef enum {
	/// Everything is fine 
	AP_SemanticMapper_OK = 0,		  	
	/// Some unspecified internal error occurred 
	AP_SemanticMapper_InternalError,   	
	/// Couldn't convert customer event to Apama event 
	AP_SemanticMapper_MappingFailure,  	
	/// Couldn't queue converted event for injection into Engine 
	AP_SemanticMapper_InjectionFailure 	
} AP_SemanticMapperError;


/**
 * AP_SemanticMapper_Functions
 *
 * Table of client visible functions exported by a Semantic Mapper instance.
 * These functions declare the only operations that may be performed by users
 * of a Semantic Mapper.
 *
 * Note that all of these functions take an initial AP_SemanticMapper*
 * argument; this is analogous to the (hidden) 'this' pointer passed to a C++
 * object when a member function is invoked on it.
 */
struct AP_SemanticMapper_Functions {

	/**
	 * sendNormalisedEvent
	 *
	 * Send a customer-specific event to the Semantic Mapper.  The event will
	 * be translated into a single Apama event that will be queued for
	 * injection into the Engine.
	 *
	 * @param mapper The Semantic Mapper to send the event to.
	 * @param event The event to send, represented as a set of name-value
	 * pairs. Ownership is transferred to the callee.
	 * @param timeStamp Timestamps associated with this event. Ownership is
	 * transferred to the callee.
	 * @return Semantic Mapper error code.  If this is not
	 * AP_SemanticMapper_OK, the getLastError() function should be called to
	 * get a more detailed description of what went wrong.
	 */
	AP_SemanticMapperError (*sendNormalisedEvent)(struct AP_SemanticMapper* mapper, AP_NormalisedEvent* event, AP_TimestampSet* timeStamp);

	/**
	 * parseApamaEvent
	 *
	 * Convert an Apama event string into one (or more) normalized events. 
	 *
	 * @param mapper The Semantic Mapper to send the event to.
	 * @param eventString The event to parse.
	 * @param channelString The channel the event was received on, or NULL
	 * @param returnArray The pointer to populate with the array of normalized events.
	 * @return Semantic Mapper error code.  If this is not
	 * AP_SemanticMapper_OK, the getLastError() function should be called to
	 * get a more detailed description of what went wrong.
	 */
	AP_SemanticMapperError (*parseApamaEvent) (struct AP_SemanticMapper* mapper, const AP_char8* eventString, const AP_char8* channelString, AP_NormalisedEvent*** returnArray);

	/**
	 * freeNormalizedArray
	 *
	 * Free a normalized event array.  Elements will NOT be deleted. 
	 *
	 * @param mapper The Semantic Mapper to send the event to.
	 * @param eventArray The array of normalized events to free (last element
	 * must be NULL).
	 * @param hasContent AP_TRUE if the array has normalized events to be deleted.
	 * @return Semantic Mapper error code.  If this is not
	 * AP_SemanticMapper_OK, the getLastError() function should be called to
	 * get a more detailed description of what went wrong.
	 */
	AP_SemanticMapperError (*freeNormalizedArray) (struct AP_SemanticMapper* mapper, AP_NormalisedEvent** eventArray, AP_bool hasContent);

	/**
	 * createApamaEventStrings
	 *
	 * Convert a customer-specific event into one (or more) Apama event
	 * strings.  Ownership of the customer-specific event is passed to
	 * to this function which will handle deleting.
	 *
	 * @param mapper The Semantic Mapper to send the event to.
	 * @param event The event to send, represented as a set of name-value
	 * pairs. Ownership is transferred to the callee.
	 * @param returnArrayPtr The pointer populate with the array of Apama events.
	 * @param channelArrayPtr The pointer populate with the array of channel names.
	 * @return Semantic Mapper error code.  If this is not
	 * AP_SemanticMapper_OK, the getLastError() function should be called to
	 * get a more detailed description of what went wrong.
	 */
	AP_SemanticMapperError (*createApamaEventStrings) (struct AP_SemanticMapper* mapper, AP_NormalisedEvent* event, AP_char8*** returnArrayPtr, AP_char8*** channelArrayPtr);

	/**
	 * freeStringArray
	 *
	 * Free an array of Apama event strings.
	 *
	 * @param mapper The Semantic Mapper to send the event to.
	 * @param eventArray The array of event strings to free (last element
	 * must be NULL).
	 * @param hasContent AP_TRUE if the array has event strings to be deleted.
	 * @return Semantic Mapper error code.  If this is not
	 * AP_SemanticMapper_OK, the getLastError() function should be called to
	 * get a more detailed description of what went wrong.
	 */
	AP_SemanticMapperError (*freeStringArray) (struct AP_SemanticMapper* mapper, AP_char8** eventArray, AP_bool hasContent);

	/**
	 * getLastError
	 *
	 * Return the Semantic Mapper's stored error message, if any.  The message
	 * string is owned by the mapper so should not be modified or freed by the
	 * caller.
	 *
	 * @param mapper The Semantic Mapper instance
	 * @return The last error message generated by the mapper
	 */
	const AP_char8* (*getLastError)(struct AP_SemanticMapper* mapper);
};


/**
 * AP_SemanticMapper
 *
 * External (client-visible) interface to the Semantic Mapper component of an
 * IAF instance.  The AP_SemanticMapper struct contains contains a table of
 * function pointers, declared in the AP_SemanticMapper_Functions struct
 * above.  The implementation of these functions is private.  Users of the
 * Semantic Mapper should invoke functions on it as in the following example
 * (mapper is of type AP_SemanticMapper*):
 *
 * i = mapper->functions->sendNormalisedEvent(mapper, event, timestamps);
 */
struct AP_SemanticMapper{

	/**
	 * Pointer to private internal data.
	 */
	void* reserved;

	/**
	 * Function table of Semantic Mapper operations.  See documentation for
	 * AP_SemanticMapper_Functions for details.
	 */
	struct AP_SemanticMapper_Functions* functions;

};
typedef struct AP_SemanticMapper AP_SemanticMapper;

/**
 * ap_getSemanticMapper
 *
 * External (client-visible) method to obtain a pointer to the Semantic Mapper 
 * component of an IAF instance.  
 */
AP_IAFCORE_API AP_SemanticMapper *ap_getSemanticMapper();

#ifdef __cplusplus
}  /* extern "C" */
#endif  /* __cplusplus */


#endif  /* AP_SEMANTIC_MAPPER_H */
